from rpc import RPCError, DefaultPoller
from rpc.common import pickle, unpickle, RemoteException
from rpc.connection import BufferedConnection
import logging
import socket
import time
import types
import traceback

CLIENT_RPCID_GEN = iter(xrange(100000000))

class ClientHandle(object):
  PENDING = 0
  SENT = 1
  SUCCESS = 2
  TIMED_OUT = 3
  ERROR = 4

  def done(self):
    return self.state >= ClientHandle.SUCCESS

  def __init__(self, rpcid, deadline = 60.0):
    self.rpcid = rpcid
    self.state = ClientHandle.PENDING
    self.start_time = time.time()
    self.server_elapsed = 0.0
    self.end_time = None
    self.result = None
    self.error = None
    self.deadline = self.start_time + deadline
#    self.finished = threading.Condition()

  def __repr__(self):
    return 'ClientHandle(rpc=%d, state=%d)' % (self.rpcid, self.state)

class ClientConnection(BufferedConnection):
  def __init__(self, host, port):
    port = int(port)
    self._callbacks = {}
    self.pending_rpcs = {}
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    while 1:
      try:
        s.connect((host, port))
        break
      except socket.error, e:
        logging.warn('Error trying to connect to %s:%d, retrying in 1s. :: %s',
                     host, port, e)
        time.sleep(1)

    BufferedConnection.__init__(self, s, (host, port))
    DefaultPoller().register(self)


  def read_finished(self):
    while self.has_message():
      rpcid, blob = self.pop_message()
      result, server_elapsed, error = unpickle(blob)
      handle = self.pending_rpcs.get(rpcid, None)
      if not handle:
        logging.info('Dropping response to stale handle %s', rpcid)
      else:
        handle.server_elapsed = server_elapsed
        handle.result = result
        handle.error = error
        handle.state = ClientHandle.SUCCESS
        handle.end_time = time.time()
        self.invoke_callback(rpcid, handle)

#      with handle.finished:
#        handle.finished.notify_all()

  def invoke_callback(self, rpcid, handle):
    if rpcid in self._callbacks:
      cb = self._callbacks[rpcid]
      cb(handle.result)
      del self._callbacks[rpcid]
      del self.pending_rpcs[rpcid]

  def write_finished(self):
#    logging.info('Write finished!')
    pass

  def send(self, method, args, timeout = 60.0):
    logging.debug('Sending RPC %s', method)
    rpcid = CLIENT_RPCID_GEN.next()
    self.pending_rpcs[rpcid] = ClientHandle(rpcid, timeout)
    self.push_message(rpcid, pickle((method, args)))
    return rpcid

  def add_callback(self, rpcid, cb):
    self._callbacks[rpcid] = cb
    if self.pending_rpcs[rpcid].state == ClientHandle.SUCCESS:
      self.invoke_callback(rpcid, self.pending_rpcs[rpcid])

  def wait_for_response(self, rpcid):
#    logging.info('Waiting...')
    if not rpcid in self.pending_rpcs:
      raise RPCError, 'Tried to wait_for_response on a non-existing message.'

    handle = self.pending_rpcs[rpcid]
    while not handle.done():
      if time.time() > handle.deadline:
        logging.warn('RPC %s (client %s, server %s:%d) timed out.',
                     handle.rpcid, socket.gethostname(), self.host, self.port)
        handle.state = ClientHandle.TIMED_OUT
        handle.end_time = time.time()

      time.sleep(0.001)
#      logging.info('Waiting... %s', self.pending_rpcs.items()[:10])
#      with handle.finished:
#        handle.finished.wait()

    del self.pending_rpcs[rpcid]

    if handle.error is not None:
      raise RPCError, handle.error

    return handle

class Future(object):
  def __init__(self, socket, rpcid):
    self.conn = socket
    self.rpcid = rpcid
    self.result = None
    self.sent = False
    self.handle = self.conn.pending_rpcs[rpcid]

  def __cmp__(self, other):
    return cmp(self.wait(), other)

  def __repr__(self):
    return 'Future for RPC(%s)' % self.rpcid

  def done(self):
    return self.handle.done()

  def start_time(self):
    return self.handle.start_time

  def end_time(self):
    return self.handle.end_time

  def server_elapsed(self):
    return self.handle.server_elapsed

  def state(self):
    return self.handle.state

  def elapsed(self):
    return self.handle.end_time - self.handle.start_time

  def set_callback(self, callback):
    self.conn.add_callback(self.rpcid, callback)

  def wait(self):
    if self.result is not None:
      return self.result

    self.sent = True
    handle = self.conn.wait_for_response(self.rpcid)
    result = handle.result
    
    # special case -- if the response only contains one argument, extract 
    # the argument and return
    if len(result) == 1: result = result[0]

    if handle.state != ClientHandle.SUCCESS:
      raise Exception('Call failed.')
    
    if isinstance(result, RemoteException):
      logging.error('Remote exception!\n: %s', result.formatted_tb)
      raise result.exception

    self.result = result
    return self.result

class _MethodCall(object):
    def __init__(self, socket, name):
      self.conn = socket
      self.name = name

    def __call__(self, *args, **kw):
      timeout = kw.get('timeout', 60)
      return Future(self.conn, self.conn.send(self.name, args, timeout = timeout))

class RPCClient(object):
  def __init__(self, host, port):
    self.host = host
    self.port = int(port)
    self.conn = ClientConnection(host, port)

  def __getattr__(self, key):
    return _MethodCall(self.conn, key)

  def close(self):
    del self.conn



from rpc import SocketError
from select import EPOLLIN, EPOLLOUT, EPOLLERR
import logging
import select
import socket
import sys
import threading

class EPollWorker(object):
  '''
  Polls a set of active socket objects, reading and writing from them
  as data is available.
'''
  def __init__(self, name = 'EPollWorker'):
    self.name = name
    self._epoll = select.epoll(500)
    self._running = True
    self._lock = threading.Lock()
    self._pending = []
    self._sockets = {}
    self._thread = threading.Thread(name = name, target = self.run)
    self._thread.setDaemon(True)
    self._thread.start()

  def stop(self):
    logging.info('Shutting down PollWorker')
    self._running = False
    self._thread.join()
    for fd, _ in self._sockets.items():
      self.unregister(fd)

  def register(self, conn):
    with self._lock:
      logging.debug('Registering %s (%d)', self.name, conn.fileno())
      self._pending.append(conn)

  def unregister(self, fd):
    with self._lock:
      logging.debug('Dropping %s (%d)', self.name, fd)
      self._epoll.unregister(fd)
      socket = self._sockets[fd]
      socket.close()
      del self._sockets[fd]

  def run(self):
    while self._running:
      try:
        self._poll_loop()
      except IOError:
        logging.debug('Ignoring IO error during poll.')
      except Exception:
        logging.warn('Error during poll loop, continuing...', exc_info = 1)

    logging.info('Killing all client connections')
    [conn.close() for conn in self._sockets.values()]
    logging.info('Client poller shutdown succesfully.')

  def _poll_loop(self):
    while self._running:
#      logging.debug('Polling %d objects...', len(self._sockets))
      events = self._epoll.poll(0.001)

      for fd, ev in events:
        try:
          conn = self._sockets[fd]
          if ev & EPOLLIN:
#            logging.info('Reading...')
            conn.read_some_bytes()
          if ev & EPOLLOUT:
#            logging.info('Writing...')
            conn.write_some_bytes()
          if ev & EPOLLERR:
            raise SocketError, 'Socket error detected.'
        except socket.error:
          self.unregister(fd)

      # update epoll and add any pending sockets
      with self._lock:
        for conn in self._sockets.values():
          if conn.has_pending_writes():
            self._epoll.modify(conn, EPOLLIN | EPOLLOUT | EPOLLERR)
          else:
            self._epoll.modify(conn, EPOLLIN | EPOLLERR)

        for conn in self._pending:
            self._epoll.register(conn, EPOLLIN | EPOLLERR)
            self._sockets[conn.fileno()] = conn

        self._pending = []



import jsonpickle
import socket
import yaml
import traceback


class RemoteException(object):
  def __init__(self, exc_info):
    self.exception = exc_info[1]
    tb  = '\n'.join(traceback.format_exception(*exc_info))
    self.formatted_tb = tb.replace('\n', '\n>>')

class Message(object):
  '''A simple helper class to build POD type objects.
  
  Object attributes can be set via keyword arguments in the constructor.'''
  def __init__(self, **kw):
    for k, v in kw.iteritems():
      setattr(self, k, v)

  def __repr__(self):
    return '%s(\n%s)' % (
      self.__class__.__name__,
      ',\n'.join(['%s : %s' % (k, v)
                  for (k, v) in self.as_dict().items()]))

  def as_dict(self):
    return dict((k, v) for k, v in self.__dict__.iteritems() if k[0] != '_')

def find_open_port():
  s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
  s.bind(("", 0))
  s.listen(1)
  port = s.getsockname()[1]
  s.close()
  return port

def split_addr(hostport):
  host, port = hostport.split(':')
  return host, int(port)


def pickle(v):
  return jsonpickle.encode(v)

def unpickle(str):
  return jsonpickle.decode(str)



#!/usr/bin/env python

from eventlet.green import httplib
from httprpc.common import Message
import logging
import types
import yaml

def load(data):
  '''Takes as input a string and returns a deserialized output object'''
  #logging.info('Loading: "%s"', data)
  return yaml.load(data)

def store(message):
  '''Takes as input an object and returns a serialized string representation.'''
  return yaml.dump(message)

class ServerRequest(Message):
  method = None
  args = None
  kw = None

class ServerResponse(Message):
  objectid = None
  data = None
  exception = None
  traceback = None

PRIMITIVES = set(
  [ types.IntType,
    types.CodeType,
    types.FloatType,
    types.BooleanType,
    types.StringType,
    types.NoneType, ])

def is_primitive(obj):
  if type(obj) in PRIMITIVES:
    return True

  if isinstance(obj, Message):
    return True

  if isinstance(obj, types.ListType):
    for v in obj:
      if not is_primitive(v):
        return False
    return True

  if isinstance(obj, types.DictType):
    for k, v in obj.iteritems():
      if not is_primitive(k) or not is_primitive(v):
        return False
    return True
  return False

class RPCError(Exception):
  pass

class ConnectionLost(RPCError):
  pass

class ServerError(RPCError):
  pass

class Channel(object):
  def __init__(self, host, port):
    self.host = host
    self.port = port

  def get(self, path, data):
    http = httplib.HTTPConnection(self.host, self.port)
    http.request('POST', path, data)
    return http.getresponse()

  def __repr__(self):
    return 'HTTP(%s:%s)' % (self.host, self.port)



from rpc.common import pickle, unpickle, RemoteException
from rpc.connection import BufferedConnection
from rpc.poller import EPollWorker
import logging
import select
import socket
import sys
import threading
import time

SERVER_RPCID_GEN = iter(xrange(100000000))

class ServerHandle(object):
  def __init__(self, connection, method, args, client_rpcid):
    self.connection = connection
    self.method = method
    self.args = args
    self.start = time.time()
    self.server_rpcid = SERVER_RPCID_GEN.next()
    self.client_rpcid = client_rpcid
    self.done = None

  def elapsed(self):
    return time.time() - self.start


class ServerConnection(BufferedConnection):
  def __init__(self, rpc_server, socket, addr):
    logging.debug('Server (%s:%d) -- accepted connection from %s',
                  rpc_server.host, rpc_server.port, addr)
    self.rpc_server = rpc_server
    BufferedConnection.__init__(self, socket, addr)

  def read_finished(self):
#    logging.info('Read finished!')
    while self.has_message():
      rpcid, message = self.pop_message()
      method, args = unpickle(message)
      self.rpc_server._dispatch(ServerHandle(self, method, args, rpcid))

  def write_finished(self):
    pass

class RPCServer(threading.Thread):
  def __init__(self, host, port, handler):
    threading.Thread.__init__(self)
    self.setDaemon(True)
    self.host = host
    self.port = port
    self.handler = handler
    self.handler.server = self
    self.running = True

    self._pending_rpcs = {}

    self._listen_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    self._listen_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    self._listen_socket.settimeout(None)
    self._listen_socket.bind((host, port))

    self._poller = EPollWorker('ServerPollWorker')

  def num_pending_rpcs(self):
    return len(self._pending_rpcs)
  
  def run(self):
    self.serve_forever()
    
  def stop(self):
    self.shutdown()

  def shutdown(self):
#    self._call_histogram.dump()
    self.running = False
    logging.debug('Shutting down!')

  def serve_forever(self):
    self._listen_socket.listen(1000)
    while self.running:
      #logging.debug('Waiting for connections...')
      s = [self._listen_socket]
      try:
        r, w, _ = select.select(s, s, s, 0.1)
      except select.error:
        continue

      if r or w:
        client_socket, addr = self._listen_socket.accept()
        conn = ServerConnection(self, client_socket, addr)
        self._poller.register(conn)

    logging.info('Listen socket shutdown successfully.')
    self._poller.stop()
    self._listen_socket.close()

  def rpc_finished(self, server_rpcid, *result):
#    logging.info('RPCFINISHED %d', server_rpcid)
    try:
      handle = self._pending_rpcs[server_rpcid]
      logging.debug('Returning result for sid %d cid %d',
                handle.server_rpcid, handle.client_rpcid)
      message = pickle((result, handle.elapsed(), None))
      handle.connection.push_message(handle.client_rpcid, message)
      del self._pending_rpcs[server_rpcid]
    except:
      logging.fatal('Unexpected error handling rpc finished!', exc_info = 1)
      time.sleep(1)

  def _dispatch(self, handle):
#    logging.info('RPCSTART %d', handle.server_rpcid)
#    logging.debug('Dispatching: %s', handle.method)
    logging.debug('Dispatching for sid %d cid %d',
                  handle.server_rpcid, handle.client_rpcid)
    self._pending_rpcs[handle.server_rpcid] = handle
    handle.done = lambda *args: self.rpc_finished(handle.server_rpcid, *args)
    try:
      getattr(self.handler, handle.method)(handle, *handle.args)
    except:
      logging.warn('Exception while handling method: %s', handle.method, exc_info = 1)
      handle.done(RemoteException(sys.exc_info()))
#    logging.debug('Dispatch finished.')




import socket
import threading

class SocketError(socket.error): pass
class RPCError(Exception): pass

_DefaultPoller = None
def DefaultPoller():
  from rpc import poller
  global _DefaultPoller
  if not _DefaultPoller:
    _DefaultPoller = poller.EPollWorker('ClientPollWorker')
  return _DefaultPoller


def ResetPoller():
  from rpc import poller
  global _DefaultPoller
  _DefaultPoller = poller.EPollWorker('ClientPollWorker')


CLIENT_CACHE = threading.local()
def connect(host, port):
  from rpc.client import RPCClient
  if not hasattr(CLIENT_CACHE, 'connections'):
    CLIENT_CACHE.connections = {}

  if not (host, port) in CLIENT_CACHE.connections:
    CLIENT_CACHE.connections[(host, port)] = RPCClient(host, port)

  client = CLIENT_CACHE.connections[(host, port)]
  return client



#!/usr/bin/env python

import os.path
README = open(os.path.abspath(os.path.dirname(__file__)) + '/README.md').read()

from setuptools import setup, find_packages, Command
setup(
    name = "speedy",
    description="Fast, non-blocking JSON based RPC system.",
    version = "0.12",
    author="Russell Power",
    author_email="power@cs.nyu.edu",
    license="BSD",
    url="http://github.com/rjpower/speedy",
    data_files = [('.', ['README.md'])],
    package_dir = { '' : 'src' },
    packages = ['rpc'],
    requires = [
      'PyYAML',
      'jsonpickle',
    ],
    install_requires = [ 
      'PyYAML',  
      'jsonpickle',
    ],
    long_description=README,
    classifiers=['Development Status :: 3 - Alpha',
                 'Topic :: Software Development :: Libraries',
                 'Topic :: System :: Clustering',
                 'Topic :: System :: Distributed Computing',
                 'License :: OSI Approved :: BSD License',
                 'Intended Audience :: Developers',
                 'Intended Audience :: System Administrators',
                 'Operating System :: POSIX',
                 'Programming Language :: Python :: 2.5',
                 'Programming Language :: Python :: 2.6',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.0',
                 'Programming Language :: Python :: 3.1',
                 'Programming Language :: Python :: 3.2',
                 ],
)



from rpc import RPCError, SocketError
import collections
import logging
import struct

class BufferedConnection(object):
  def __init__(self, socket, addr):
    self.host = addr[0]
    self.port = addr[1]
    self._socket = socket
    self._socket.setblocking(False)
    self._pending_writes = collections.deque()
    self._pending_reads = collections.deque()
    self._read_buffer = ''

  def __repr__(self):
    return repr(self._socket)

  def close(self):
    self._socket.close()

  def fileno(self):
    return self._socket.fileno()

  def read_finished(self):
    '''Invoked whenever any data has been read.'''
    raise NotImplementedError

  def write_finished(self):
    '''Invoked whenever any data has been written.'''
    raise NotImplementedError

  def has_pending_writes(self):
    return self._pending_writes

  def push_message(self, rpc_id, message):
    header = struct.pack('II', len(message), rpc_id)
    self._pending_writes.append((rpc_id, header + message))

  def pop_message(self):
    return self._pending_reads.popleft()

  def has_message(self):
    return self._pending_reads

  def read_some_bytes(self):
    in_bytes = self._socket.recv(100000)
    self._read_buffer = self._read_buffer + in_bytes

    # A read size of 0 is used to detect a connection close.
    if len(in_bytes) == 0:
      raise SocketError, 'Connection closed.'

    while len(self._read_buffer) > 8:
      (payload_len, rpcid) = struct.unpack('II', self._read_buffer[:8])
      if len(self._read_buffer) >= 8 + payload_len:
        payload = self._read_buffer[8:8 + payload_len]
        self._pending_reads.append((rpcid, payload))
        self._read_buffer = self._read_buffer[8 + payload_len:]
      else:
        break

#    logging.info('Read %d bytes', len(in_bytes))
    self.read_finished()
    return len(in_bytes)

  def write_some_bytes(self):
    if not self._pending_writes:
      return 0

    rpcid, write_buffer = self._pending_writes.popleft()
    bytes_written = self._socket.send(write_buffer)
    if bytes_written != len(write_buffer):
      self._pending_writes.appendleft((rpcid, write_buffer[bytes_written:]))
      self.write_finished()
    else:
      pass
#      logging.info('Finished send of rpc %s', rpcid)

#    logging.info('Wrote %d bytes', bytes_written)
    return bytes_written
